/* Copyright (c) 2009 The Regents of the Signet Group.
 *  All rights reserved.
 *
 *  Permission to use, copy, modify, and distribute this software and its
 *  documentation for any purpose, without fee, and without written
 *  agreement is hereby granted, provided that the above copyright
 *  notice, the (updated) modification history and the author appear in
 *  all copies of this source code.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS `AS IS'
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS
 *  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, LOSS OF USE, DATA,
 *  OR PROFITS) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *  THE POSSIBILITY OF SUCH DAMAGE.
 *  
 *  Authors:
 *  Angelo P. Castellani <castellani@dei.unipd.it>
 *  Mattia Gheda <ghedamat@dei.unipd.it>
 */

#ifndef COAP_H
#define COAP_H

#ifndef COAP_MAX_TRANSACTIONS
#define COAP_MAX_TRANSACTIONS 5
#endif

#define COAP_UNDEF_TID 0


#define LOWPAN_XPORT_START 61616u
#define LOWPAN_CLIENT_PORT_START 61618u


#ifndef COAP_MAX_PKT_LEN
#define COAP_MAX_PKT_LEN 300
#endif

#ifndef COAP_MAX_URI_LEN
#define COAP_MAX_URI_LEN 100
#endif

#include "IP.h"

typedef uint8_t coap_contenttype_t;
typedef uint8_t coap_char_t;
typedef uint8_t coap_length_t;
typedef uint8_t coap_status_t;
typedef uint8_t coap_method_t;
typedef coap_char_t coap_uri_t;
typedef uint16_t coap_sid_t;
typedef uint8_t coap_cid_t;
typedef uint16_t coap_tid_t;
typedef uint8_t coap_period_t;
typedef uint8_t coap_msgtype_t;

typedef struct {
	ip6_addr_t* host;
	uint16_t port;
	coap_uri_t *uri;
} coap_absuri_t;

typedef struct {
	coap_char_t* data;
	coap_length_t len;
	coap_contenttype_t format;
} coap_content_t;

typedef struct {
	coap_tid_t tid;
	coap_sid_t sid;
	ip6_addr_t host;
	uint16_t port;
	coap_char_t pkt[COAP_MAX_PKT_LEN];
	coap_length_t len;
	bool ack; // Waiting for an ACK
	bool client;
#ifdef COAP_MEASURE_SERVICE_TIME
	uint32_t time;
#endif
} coap_transaction_t;


typedef coap_transaction_t* coap_rid_t; // request ID Identifies a request

typedef struct {
	coap_char_t* buf;
	coap_length_t len;
	coap_tid_t tid;
	coap_msgtype_t msgtype;
} coap_packet_t;

typedef struct {
	coap_method_t method;
	coap_status_t status;
	coap_absuri_t* absuri;
	coap_content_t* content;
	bool ack;
#ifdef COAP
#ifdef COAP_BLOCK_OPTION
	bool chunked;
	uint32_t bn;
	uint8_t szx;
	bool mb;
#endif 
#endif
} coap_msg_t;


#ifndef COAP_TIMEOUT
#define COAP_TIMEOUT 300
#endif

#ifndef COAP_NOACK_TIMEOUTS
#define COAP_NOACK_TIMEOUTS 5
#endif

#ifndef COAP_MAX_RETRANSMIT
#define COAP_MAX_RETRANSMIT 5
#endif 

void write_packet( coap_packet_t* pkt, coap_msg_t* data);
void parse_packet( coap_packet_t* pkt, coap_msg_t* data);

#ifndef COAP
#include "BWS-08.h"
#else
#include "CoAP-01.h"
#endif

#endif
