/* Copyright (c) 2009 The Regents of the Signet Group.
*  All rights reserved.
*
*  Permission to use, copy, modify, and distribute this software and its
*  documentation for any purpose, without fee, and without written
*  agreement is hereby granted, provided that the above copyright
*  notice, the (updated) modification history and the author appear in
*  all copies of this source code.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS `AS IS'
*  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
*  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
*  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS
*  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
*  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, LOSS OF USE, DATA,
*  OR PROFITS) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
*  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
*  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
*  THE POSSIBILITY OF SUCH DAMAGE.
*
*  Authors:
*  Angelo P. Castellani <castellani@dei.unipd.it>
*  Mattia Gheda <ghedamat@dei.unipd.it>
*/
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

typedef uint8_t bool;
typedef unsigned char nx_uint8_t;
typedef uint16_t nx_uint16_t;
typedef uint32_t nx_uint32_t;

#define TRUE 1
#define FALSE 0

//#define BWS
#define COAP

#ifdef COAP
#define COAP_BLOCK_OPTION
#endif

#include "CoAP.h"

#ifdef COAP_BLOCK_OPTION
uint32_t bn = 0;
uint8_t szx = 0;
bool more = FALSE;
#endif

uint16_t tid = 0;

void send_packet(char* b, int* len, coap_uri_t* uri) {
	coap_packet_t pkt;
	coap_msg_t msg;

	coap_absuri_t absuri;
	//coap_uri_t uri[100] = "/";
	coap_char_t cdata[100];
	coap_content_t content;
	
	pkt.buf = b;
	pkt.msgtype = COAP_MSGTYPE_REQUEST;
	pkt.len = *len;
	pkt.tid = tid++;

	absuri.uri = uri;
	content.data = cdata;

	msg.absuri = &absuri;
	msg.content = NULL;
	msg.method = COAP_METHOD_GET;
	msg.ack = TRUE;

#ifdef COAP_BLOCK_OPTION	
	msg.chunked = FALSE;
	if (more) {
		msg.chunked = TRUE;
		msg.mb = FALSE;
		msg.bn = bn;
		msg.szx = szx;
	}
#endif

	write_packet(&pkt, &msg);

	*len = pkt.len;
}

#define BUF_SIZE 1024

int main (int argc, char** argv) {
	int s = socket( AF_INET6, SOCK_DGRAM, IPPROTO_UDP );
	struct sockaddr_in6 sa;
	char b[BUF_SIZE];
	ssize_t recsize;
	socklen_t fromlen;
	int sent_len, blen = BUF_SIZE;
	int i;

	if (argc<5) {
		fprintf(stderr,"not enought parameters..\n");
		printf("USAGE:  ./coap 2001:638:709:1234::fffe:2 61618 RESOURCE NUM_OF_REQUESTS\n");
		close(s);
		exit(EXIT_FAILURE);
	}

	memset( &sa, 0, sizeof(sa) );

	sa.sin6_family = AF_INET6;
	sa.sin6_addr = in6addr_any;
	sa.sin6_port = htons(61616);

	if (-1 == bind(s, (struct sockaddr *)&sa, sizeof(struct sockaddr_in6) ) ) {
		perror("error during bind");
		close(s);
		exit(EXIT_FAILURE);
	}

	if (!inet_pton( AF_INET6, argv[1], &sa.sin6_addr )) {
		perror("error on parameter 1");
		close(s);
		exit(EXIT_FAILURE);
	}
	sa.sin6_port = htons(atoi(argv[2]));

	
	for (i=0; i<atoi(argv[4]); i++) {
		send_packet(b, &blen, argv[3]);
		sent_len = sendto(s, b, blen, 0,(struct sockaddr*)&sa, sizeof (struct sockaddr_in6));
		usleep(200);
	}

	while (1) {
		char* p = b;

		recsize = recvfrom(s, (void *)b, BUF_SIZE, 0, (struct sockaddr *)&sa, &fromlen);

		if (recsize > 0) {
			coap_content_t content;
			coap_absuri_t absuri;
			coap_uri_t uri[100];
			coap_msg_t msg;
			coap_packet_t pkt;
			coap_char_t data[1000];

			pkt.buf = b;
			pkt.len = recsize;

			msg.absuri = &absuri;
			absuri.uri = uri;
			msg.content = &content;

			parse_packet( &pkt, &msg );

			strncpy(data,content.data,content.len);

			printf("status:%u content->format:%u tid:%u\n", msg.status, content.format, pkt.tid );
			printf("content.data:\n%s\n\n", data);

			printf("hex dump of RESPONSE:\n");
			while (content.len--) { printf("%02x ", (unsigned int)*(content.data++)&0xff); }
			printf("\n\n");

#ifdef COAP_BLOCK_OPTION			
			if (msg.chunked && msg.mb) {
				
				blen = BUF_SIZE;
				more = TRUE;
				bn = msg.bn + 1;
				szx = msg.szx;
				
				printf("requesting block #%u\n", bn);
				
				send_packet(b, &blen, argv[3]);	

				sent_len = sendto(s, b, blen, 0,(struct sockaddr*)&sa, sizeof (struct sockaddr_in6));
				
			} else more = FALSE;
#endif
		}
	}

	close(s);
}
